from zipfile import ZipFile
import tempfile
from os import path

_shp_exts = ["dbf","prj","shx"]
_shp_exts = _shp_exts + map(lambda s: s.upper(), _shp_exts)

def shp_files(fpath):
    basename, ext = path.splitext(fpath)
    paths = [ "%s.%s" % (basename,ext) for ext in _shp_exts ]
    paths.append(fpath)
    return filter(lambda f: path.exists(f),  paths)
    
def create_zip(fpaths):
    _,payload = tempfile.mkstemp(suffix='.zip')
    zipf = ZipFile(payload, "w")
    for fp in fpaths:
            basename = path.basename(fp)
            zipf.write(fp, basename)
    zipf.close()
    return payload



import httplib2
import logging
from gsimporter.api import parse_response
from gsimporter.api import RequestFailed
from gsimporter.api import BadRequest
from gsimporter.api import NotFound
from urlparse import urlparse
from urllib import urlencode
import os
import _util
import pprint
import json
import mimetypes

_logger = logging.getLogger(__name__)

class Client(object):

    def __init__(self, url, username=None, password=None):
        username = username or 'admin'
        password = password or 'geoserver'
        self.client = _Client(url,username,password)

    def _call(self, fun, *args):
        # call the provided function and set the _uploader field on each
        # of the returned api objects
        robj = fun(*args)
        if isinstance(robj, list):
            for i in robj:
                i._uploader = self
        else:
            robj._uploader = self
        return robj

    def get_sessions(self):
        '''Get an 'unexpanded' list of a imports - the Session objects returned
        will not have any details beside their id and href.
        :return a list of Session objects
        '''
        return self._call(self.client.get_imports)
        
    def get_session(self,id):
        '''Get an existing 'expanded' session by id.
        :param id: the integer id of the session to get
        :return: an expanded Session object
        '''
        return self._call(self.client.get_import,id)

    def start_import(self, import_id=None, mosaic=False, name=None, target_store=None):
        """Create a new import session.
        :param import_id: optional id to specify
        :param mosaic: if True, indicates a mosaic upload
        :param name: if provided, name the mosaic upload
        :param target_store: if provided, name of an existing store to be updated
        :returns: a gsimporter.api.Session object
        """
        session = self._call(self.client.start_import, import_id, mosaic, name, target_store)
        if import_id: assert session.id >= import_id
        return session
        
    def upload(self, fpath, use_url=False, import_id=None, mosaic=False,
               initial_opts=None):
        """Try a complete import - create a session and upload the provided file.
        fpath can be a path to a zip file or the 'main' file if a shapefile or
        a tiff.
        :param fpath: path to file or zip
        :param use_url: if True, tell the importer where to find the file
        :param import_id: if provided, use as the specified id
        :param mosaic: if True, indicates a mosaic upload
        :param initial_opts: default None, dict of initial import options
        :returns: a gsimporter.api.Session object
        """
        files = [ fpath ]
        if fpath.lower().endswith(".shp"):
            files = _util.shp_files(fpath)

        return self.upload_files(files, use_url, import_id, mosaic, initial_opts)

    def upload_files(self, files, use_url=False, import_id=None, mosaic=False,
                     intial_opts=None, target_store=None):
        """Upload the provided files. If a mosaic, compute a name from the
        provided files.
        :param files: the files to upload
        :param use_url: if True, tell the importer where to find the file
        :param import_id: if provided, use as the specified id
        :param mosaic: if True, indicates a mosaic upload
        :param initial_opts: default None, dict of initial import options
        :returns: a gsimporter.api.Session object
        """
        name = None
        if mosaic:
            # @hack - ensure that target layer gets a nice name
            layername = os.path.basename(files[0])
            name, _ = os.path.splitext(layername)
        if target_store:
            name = os.path.abspath(files[0])

        session = self.start_import(import_id, mosaic=mosaic, name=name, target_store=target_store)
        session.upload_task(files, use_url, intial_opts)
        
        return session

        
    # pickle protocol - client object cannot be serialized
    # this allows api objects to be seamlessly pickled and loaded without restarting
    # the connection more explicitly but this will have consequences if other state is stored
    # in the uploader or client objects
    def __getstate__(self):
        cl = self.client
        return {'url':cl.service_url,'username':cl.username,'password':cl.password}
    def __setstate__(self,state):
        self.client = _Client(state['url'],state['username'],state['password'])
        
        
class _Client(object):
    """Lower level http client"""

    def __init__(self, url, username, password):
        self.service_url = url
        if self.service_url.endswith("/"):
            self.service_url = self.service_url.strip("/")
        self.http = httplib2.Http()
        self.username = username
        self.password = password
        self.http.add_credentials(self.username, self.password)
        netloc = urlparse(url).netloc
        self.http.authorizations.append(
            httplib2.BasicAuthentication(
                (username, password),
                netloc,
                url,
                {},
                None,
                None,
                self.http
            ))
            
    def url(self,path):
        return "%s/%s" % (self.service_url,path)

    def post(self, url):
        return self._request(url, "POST")
    
    def delete(self, url):
        return self._request(url, "DELETE")
        
    def put_json(self, url, data):
        return self._request(url, "PUT", data, {
            "Content-type" : "application/json"
        })
        
    def _parse_errors(self, content):
        try:
            resp = json.loads(content)
        except ValueError:
            return [content]
        return resp['errors']
    
    def _request(self, url, method="GET", data=None, headers={}):
        log_data = None
        if data:
            try:
                data_len = len(data)
            except:
                data_len = -1
            log_data = data if data_len < 1024 else "[Data to long...]"
        _logger.info("%s request to %s:\n[Data]:%s", method, url, log_data)
        resp, content = self.http.request(url, method, data, headers)
        _debug(resp, content)
        if resp.status == 404:
            raise NotFound()
        if resp.status < 200 or resp.status > 299:
            if resp.status == 400:
                raise BadRequest(*self._parse_errors(content))
            raise RequestFailed(resp.status,content)
        return resp, content
    
    def post_upload_url(self, url, upload_url):
        data = urlencode({
            'url' : upload_url
        })
        return self._request(url, "POST", data, {
            # importer very picky
            'Content-type' : "application/x-www-form-urlencoded"
        })
        
    def put_zip(self,url,payload):
        message = open(payload)
        with message:
            return self._request(url,"PUT",message,{
                "Content-type": "application/zip",
            })
            
    def get_import(self,i):
        return parse_response(self._request(self.url("imports/%s?expand=3" % i)))

    def get_imports(self):
        return parse_response(self._request(self.url("imports")))
    
    def start_import(self, import_id=None, mosaic=False, name=None, target_store=False):
        method = 'POST'
        data = None
        headers = {}
        if mosaic:
            data = json.dumps({"import": {
                "data": {
                   "type": "mosaic",
                   "name": name,
                   "time": {
                        "mode": "auto"
                   }
                }
            }})
            headers["Content-type"] = "application/json"
        if target_store:
            data = json.dumps({"import": {
                "data": {
                    "type": "file",
                    "file": name
                },
                "targetStore": {
                   "dataStore": {
                        "name": target_store
                   }
                }
            }})
            headers["Content-type"] = "application/json"
        if import_id is not None:
            url = self.url("imports/%s" % import_id)
            method = 'PUT'
        else:
            url = self.url("imports")
        return parse_response(self._request(url, method, data, headers))
        
    def post_multipart(self,url,files,fields=[]):
        """
        fields is a sequence of (name, value) elements for regular form fields.
        files is a sequence of name or (name,filename) or (name, filename, value) 
        elements for data to be uploaded as files
        
        """
        BOUNDARY = '----------ThIs_Is_tHe_bouNdaRY_$'
        CRLF = '\r\n'
        L = []
        _logger.info("post_multipart %s %s %s",url,files,fields)
        for (key, value) in fields:
            L.append('--' + BOUNDARY)
            L.append('Content-Disposition: form-data; name="%s"' % str(key))
            L.append('')
            L.append(str(value))
        for fpair in files:
            if isinstance(fpair,basestring):
                fpair = (fpair,fpair)
            key = fpair[0]
            if len(fpair) == 2:
                filename = os.path.basename(fpair[1])
                fp = open(fpair[1], 'rb')
                value = fp.read()
                fp.close()
            else:
                filename, value = fpair[1:]
            L.append('--' + BOUNDARY)
            L.append('Content-Disposition: form-data; name="%s"; filename="%s"' % (str(key), str(filename)))
            L.append('Content-Type: %s' % str(_get_content_type(filename)))
            L.append('')
            L.append(value)
        L.append('--' + BOUNDARY + '--')
        L.append('')
        return self._request(
            url, 'POST', CRLF.join(L), {
                'Content-Type' : 'multipart/form-data; boundary=%s' % BOUNDARY
            }
        )
        
        
def _get_content_type(filename):
    return mimetypes.guess_type(filename)[0] or 'application/octet-stream'

def _debug(resp, content):
    if _logger.isEnabledFor(logging.DEBUG):
        _logger.debug("response : %s",pprint.pformat(resp))
        if "content-type" in resp and resp['content-type'] == 'application/json':
            try:
                content = json.loads(content) 
                content = json.dumps(content,indent=2)
            except ValueError:
                pass

        _logger.debug("content : %s",content)



import collections
import logging
import json
import os
import pprint

STATE_PENDING = "PENDING"
STATE_READY = "READY"
STATE_RUNNING = "RUNNING"
STATE_INCOMPLETE = "INCOMPLETE"
STATE_COMPLETE = "COMPLETE"

_logger = logging.getLogger("gsuploader")

class BadRequest(Exception):
    '''Encapsulate a 400 or other 'invalid' request'''
    pass

class RequestFailed(Exception):
    '''Encapsulate a 500 or other 'failed' request'''
    pass

class NotFound(Exception):
    '''Encapsulate an HTTP 404 or other 'missing' request'''
    pass

class BindingFailed(Exception):
    '''Something in the API has changed'''
    pass

def parse_response(args, parent=None):
    '''Parse a top-level concept from the provided args.
    :param args: a tuple of headers, response from an httplib2 request
    :param parent: if present, set the parent of any returned object
    :returns: a Session, Task or list of Session or Task
    '''
    headers, response = args
    try:
        resp = json.loads(response)
    except ValueError,ex:
        _logger.warn('invalid JSON response: %s',response)
        raise ex
    if "import" in resp:
        return Session(json=resp['import'], parent=parent)
    elif "task" in resp:
        return Task(resp['task'], parent=parent)
    elif "imports" in resp:
        return [ Session(json=j, parent=parent) for j in resp['imports'] ]
    elif "tasks" in resp:
        # non-recognized file tasks have null source.format
        return [ Task(t, parent=parent) for t in resp['tasks'] ]
    raise Exception("Unknown response %s" % resp)


_Binding = collections.namedtuple('Binding', [
        'name', 'expected', 'ro', 'binding'
    ])
def _binding(name, expected=True, ro=True, binding=None):
    return _Binding(name, expected, ro, binding)



class _UploadBase(object):
    _object_name = None

    def __init__(self, json, parent=None):
        self._parent = parent
        if parent == self:
            raise Exception('bogus')
        if json is not None:
            self._bind_json(json)
        self._vals = {}
        self._uploader = None

    def reload(self):
        '''reset the state of this object if possible'''
        if not hasattr(self, 'href'):
            raise Exception('cannot reload %s', type(self))
        resp = self._client()._request(self.href + "?expand=3")
        # if we have a parent, keep it, otherwise, we're the new parent
        parsed = parse_response(resp, parent=self._parent or self)
        for binding in self._bindings:
            setattr(self, binding.name, getattr(parsed, binding.name))
        return self

    def save(self, deep=False):
        raise Exception('not implemented')

    def _bind_json(self, json):
        # generally, not required for override. instead use _bind_custom_json
        # if possible
        if not isinstance(json, dict):
            self._binding_failed('expected dict, got %s', type(json))
        for binding in self._bindings:
            val = json.pop(binding.name, None)
            if binding.expected and val is None:
                self._binding_failed('expected val for %s', binding.name)
            if binding.binding and val is not None:
                if isinstance(val, list):
                    val = [binding.binding(v, parent=self) for v in val]
                else:
                    val = binding.binding(val, parent=self)
            setattr(self, binding.name, val)
        self._bind_custom_json(json)

    def _bind_custom_json(self, json):
        # do any custom binding like for a shortcut
        pass

    def _binding_failed(self, msg, args):
        raise BindingFailed('[%s] %s' % (type(self), msg % args))

    def _getuploader(self):
        comp = self
        while comp:
            if comp._uploader:
                return comp._uploader
            comp = comp._parent

    def _url(self,spec,*parts):
        return self._getuploader().client.url( spec % parts )

    def _client(self):
        return self._getuploader().client

    def _to_json_object(self, deep=True, top_level=True):
        json = {}
        for binding in self._bindings:
            val = getattr(self, binding.name, None)
            if isinstance(val, _UploadBase):
                val = val._to_json_object(top_level=False)
            if val is not None:
                json[binding.name] = val
        self._to_json_object_custom(json)
        if top_level and self._object_name:
            json = { self._object_name : json }
        return json

    def _to_json_object_custom(self, json):
        pass

    def __repr__(self):
        jsonobj = self._to_json_object(deep=True)
        return pprint.pformat(jsonobj, indent=4)


class Data(_UploadBase):
    _object_name = 'data'
    _bindings = (
        _binding('type'),
        # a missing format indicates an undetected format
        _binding('format', expected=False),
        _binding('charset', expected=False),
        # either file or files...
        _binding('file', expected=False),
        _binding('files', expected=False),
        # location is used instead of file for remote uploads
        # e.g.:
        #   "data": {
        #       "type": "remote",
        #       "location": "ftp://myserver/data/bc_shapefiles",
        #       "username": "dan",
        #       "password": "secret"
        #   }
        _binding('location', expected=False),
        _binding('username', expected=False),
        _binding('password', expected=False),
    )


class Target(_UploadBase):
    '''This is basically a StoreInfo'''
    _store_types = ('dataStore', 'coverageStore', 'store')
    _object_name = 'target'
    _bindings = (
        _binding('name'),
        _binding('type'),
        _binding('enabled', expected=False),
    )

    def _bind_json(self, json):
        if json.has_key('href'):
            self.href = json.get('href')
        store_type = [ k for k in Target._store_types  if k in json]
        if store_type:
            if len(store_type) != 1:
                self.binding_failed('invalid store entry: %s', json.keys())
            else:
                self.store_type = store_type[0]
                repr = json[self.store_type]
                super(Target, self)._bind_json(repr)

    def _bind_custom_json(self, json):
        workspace = json.get('workspace', None)
        if workspace:
            self.workspace_name = workspace['name']

    def change_datastore(self, store_name=None, workspace=None):
        '''Immediately change the target to the specified datastore. The workspace
        must exist prior to setting as the target.

        :param store_name: An optional existing datastore name
        :param workspace: An optional workspace to use for referencing the store
        '''
        dataStore = { 'enabled' : True } # workaround for importer bug
        if store_name:
            dataStore['name'] = store_name
        if workspace:
            dataStore['workspace'] = { 'name' : str(workspace) }
        target_rep = { self.store_type : dataStore }
        self._client().put_json(self.href,json.dumps(target_rep))


class BBox(_UploadBase):
    _bindings = (
        _binding('minx'),
        _binding('miny'),
        _binding('maxx'),
        _binding('maxy'),
        _binding('crs', expected=False),
    )


class Attribute(_UploadBase):
    _bindings = (
        _binding('name'),
        _binding('binding')
    )


class Layer(_UploadBase):
    _object_name = 'layer'
    _bindings = (
        _binding('name'),
        _binding('href'),
        _binding('originalName', expected=False),
        _binding('nativeName', expected=False),
        _binding('srs', expected=False),
        _binding('attributes', binding=Attribute, expected=False),
        _binding('bbox', binding=BBox),
    )

    def set_target_layer_name(self, name):
        data = { 'layer' : { 'name' : name }}
        self._client().put_json(self.href, json.dumps(data))


class Task(_UploadBase):
    _object_name = 'task'
    _bindings = (
        _binding('id'),
        _binding('href'),
        _binding('state'),
        _binding('progress'),
        _binding('updateMode', expected=False), # workaround for older versions
        _binding('data', binding=Data),
        # a missing target implies the source must be imported into db
        _binding('target', binding=Target, expected=False),
        # a missing layer probably indicates an undetected format
        _binding('layer', binding=Layer, expected=False),
    )

    def _bind_custom_json(self, json):
        self.error_message = json.get('errorMessage', None)
        self.transform_type = json['transformChain']['type']
        self.transforms = json['transformChain']['transforms']

    # Convience methods
    def get_target_layer_name(self):
        '''Get the workspace prefixed target layer name.
        For example: `topp:world_borders`
        '''
        return '%s:%s' % (self.target.workspace_name, self.layer.name)

    # Mutators
    def set_target(self, store_name=None, workspace=None, store_type=None):
        if self.target is None:
            if workspace is None:
                raise Exception("workspace required if target is not set")
            if store_type not in Target._store_types:
                raise Exception("store_type must be one of %s" % (Target._store_types,))
            self.target = Target(None, self)
            self.target.store_type = store_type
            self.target.href = self.href + "/target"
        self.target.change_datastore(store_name, workspace)

    def set_update_mode(self,update_mode):
        data = { 'task' : {
            'updateMode' : update_mode
        }}
        self._client().put_json(self.href,json.dumps(data))

    def set_charset(self,charset):
        data = { 'task' : {
            'source' : {
                'charset' : charset
            }
        }}
        self._client().put_json(self.href,json.dumps(data))

    def set_srs(self, srs):
        data = { 'layer' : { 'srs' : srs }}
        self._client().put_json(self.href, json.dumps(data))

    def delete(self):
        """Delete the task"""
        resp, content = self._client().delete(self.href)
        if resp['status'] != '204':
            raise Exception('expected 204 response code, got %s' % resp['status'],content)

    def set_transforms(self, transforms, save=True):
        """Set the transforms of this Item. transforms is a list of dicts"""
        self.transforms = list(transforms)
        save and self.save_transforms()

    def save_transforms(self):
        chain = {
            "type": self.transform_type,
            "transforms": self.transforms
        }
        data = { 'task' : { 'transformChain' : chain}}
        self._client().put_json(self.href, json.dumps(data))

    def add_transforms(self, transforms, save=True):
        self.transforms.extend(transforms)
        save and self.save_transforms()

    def remove_transforms(self, transforms, by_field=None, save=True):
        '''remove transforms by equality or list of field values'''
        if by_field:
            self.transforms = [ t for t in self.transforms if t[by_field] not in transforms ]
        else:
            self.transforms = [ t for t in self.transforms if t not in transforms ]
        save and self.save_transforms()

    def get_progress(self):
        """Get a json object representing progress of this item"""
        if self.progress:
            client = self._client()
            headers, response = client._request(self.progress)
            unicode_error = False
            try:
                response = response.decode('utf-8', 'strict')
            except UnicodeDecodeError:
                response = response.decode('utf-8', 'replace')
                unicode_error = True
            try:
                progress = json.loads(response)
                if unicode_error:
                    progress['message'] += ' - it looks like an invalid character'
                return progress
            except ValueError,ex:
                _logger.warn('invalid JSON response: %s',response)
                raise RequestFailed('invalid JSON')
        else:
            raise Exception("Item does not have a progress endpoint")


class Session(_UploadBase):
    _object_name = 'import'
    _bindings = (
        _binding("id"),
        _binding("href"),
        _binding("state"),
        _binding("data", binding=Data, expected=False),
        _binding("archive", expected=False, ro=False),
        _binding("tasks", expected=False, binding=Task),
        _binding("targetWorkspace", expected=False, binding=Target),
        _binding("targetStore", expected=False, binding=Target),
    )

    def delete_unrecognized_tasks(self):
        """For any tasks that have unrecognized formats, wipe them out."""
        for t in list(self.tasks):
            if t.data.format is None:
                _logger.info("deleting unrecognized task %s", t)
                t.delete()
                self.tasks.remove(t)

    def upload_task(self, files, use_url=False, initial_opts=None):
        """create a task with the provided files
        files - collection of files to upload or zip file
        use_url - if true, post a URL to the uploader
        """
        # @todo getting the task response updates the session tasks, but
        # neglects to retreive the overall session status field
        fname = os.path.basename(files[0])
        _,ext = os.path.splitext(fname)
        def addopts(base):
            if initial_opts:
                # pass options in as value:key parameters, this allows multiple
                # options per key
                base = base + '&' + '&'.join(['option=%s:%s' % (v,k) for k,v in initial_opts.iteritems()])
            return base
        if use_url:
            if ext == '.zip':
                upload_url = files[0]
            else:
                upload_url = os.path.dirname(files[0])
            url = self._url("imports/%s/tasks?expand=3" % self.id)
            upload_url = "file://%s" % os.path.abspath(upload_url)
            resp = self._client().post_upload_url(url, upload_url)
        elif ext == '.zip':
            url = self._url("imports/%s/tasks/%s?expand=3" % (self.id, fname))
            resp = self._client().put_zip(addopts(url), files[0])
        else:
            url = self._url("imports/%s/tasks?expand=3" % self.id)
            resp = self._client().post_multipart(addopts(url), files)
        tasks = parse_response( resp )
        if not isinstance(tasks, list):
            tasks = [tasks]
        for t in tasks:
            t._parent = self
        self.tasks.extend(tasks)

    def commit(self, async=False):
        """Run the session"""
        #@todo check task status if we don't have it already?
        url = self._url("imports/%s",self.id)
        if async:
            url = url + "?async=true&exec=true"
        resp, content = self._client().post(url)
        if resp['status'] != '204':
            raise Exception("expected 204 response code, got %s" % resp['status'],content)

    def delete(self):
        """Delete this import session"""
        url = self._url("imports/%s",self.id)
        resp, content = self._client().delete(url)
        if resp['status'] != '204':
            raise Exception('expected 204 response code, got %s' % resp['status'],content)



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(name = "gn_gsimporter",
    version = "1.0.2",
    description = "GeoNode GeoServer Importer Client",
    keywords = "GeoNode GeoServer Importer",
    license = "MIT",
    url = "https://github.com/GeoNode/gsimporter",
    author = "Ian Schneider",
    author_email = "ischneider@opengeo.org",
    install_requires = [
        'httplib2',
    ],
    tests_require = [
        'gisdata>=0.5.4',
        'gsconfig>=1.0.0',
        'psycopg2',
        'OWSLib>=0.7.2',
        'unittest2',
    ],
    package_dir = {'':'src'},
    packages = find_packages('src'),
    test_suite = 'test.uploadtests'
)
